from typing import Any, Optional, Literal

from pydantic import BaseModel, ConfigDict

from phi.assistant.exceptions import FileIdNotSet
from phi.utils.log import logger

try:
    from openai import OpenAI
    from openai.types.file_object import FileObject as OpenAIFile
    from openai.types.file_deleted import FileDeleted as OpenAIFileDeleted
except ImportError:
    logger.error("`openai` not installed")
    raise


class File(BaseModel):
    # -*- File settings
    # File id which can be referenced in API endpoints.
    id: Optional[str] = None
    # The object type, populated by the API. Always file.
    object: Optional[str] = None

    # The size of the file, in bytes.
    bytes: Optional[int] = None

    # The name of the file.
    filename: Optional[str] = None
    # The intended purpose of the file.
    # Supported values are fine-tune, fine-tune-results, assistants, and assistants_output.
    purpose: Literal["fine-tune", "assistants"] = "assistants"

    # The Unix timestamp (in seconds) for when the file was created.
    created_at: Optional[int] = None

    openai: Optional[OpenAI] = None
    openai_file: Optional[OpenAIFile] = None

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @property
    def client(self) -> OpenAI:
        return self.openai or OpenAI()

    def read(self) -> Any:
        raise NotImplementedError

    def load_from_openai(self, openai_file: OpenAIFile):
        self.id = openai_file.id
        self.object = openai_file.object
        self.created_at = openai_file.created_at

    def upload(self) -> OpenAIFile:
        self.openai_file = self.client.files.create(file=self.read(), purpose=self.purpose)
        self.load_from_openai(self.openai_file)
        return self.openai_file

    def download(self, use_cache: bool = True) -> str:
        try:
            file_to_download = self.get(use_cache=use_cache)
            if file_to_download is not None:
                content = self.client.files.retrieve_content(file_id=file_to_download.id)
                return content
        except FileIdNotSet:
            logger.warning("File not available")
            raise

    def get(self, use_cache: bool = True) -> OpenAIFile:
        if self.openai_file is not None and use_cache:
            return self.openai_file

        _file_id = self.id or self.openai_file.id if self.openai_file else None
        if _file_id is not None:
            self.openai_file = self.client.files.retrieve(file_id=_file_id)
            self.load_from_openai(self.openai_file)
            return self.openai_file
        raise FileIdNotSet("File.id not set")

    def delete(self) -> OpenAIFileDeleted:
        try:
            file_to_delete = self.get()
            if file_to_delete is not None:
                deletion_status = self.client.files.delete(
                    file_id=file_to_delete.id,
                )
                return deletion_status
        except FileIdNotSet:
            logger.warning("File not available")
            raise



from typing import Optional, Union, List

from phi.k8s.app.airflow.base import AirflowBase


class AirflowWebserver(AirflowBase):
    # -*- App Name
    name: str = "airflow-ws"

    # Command for the container
    command: Optional[Union[str, List[str]]] = "webserver"

    # -*- App Ports
    # Open a container port if open_port=True
    open_port: bool = True
    port_number: int = 8080

    # -*- Service Configuration
    create_service: bool = True



from typing import List, Any, Optional, Dict

from pydantic import BaseModel, ConfigDict

from phi.assistant.exceptions import ThreadIdNotSet, MessageIdNotSet
from phi.utils.log import logger

try:
    from openai import OpenAI
    from openai.types.beta.threads.thread_message import ThreadMessage as OpenAIThreadMessage
except ImportError:
    logger.error("`openai` not installed")
    raise


class Message(BaseModel):
    # -*- Message settings
    # Message id which can be referenced in API endpoints.
    id: Optional[str] = None
    # The object type, populated by the API. Always thread.message.
    object: Optional[str] = None

    # The entity that produced the message. One of user or assistant.
    role: Optional[str] = None
    # The content of the message in array of text and/or images.
    content: List[Any] | str

    # The thread ID that this message belongs to.
    # Required to create/get a message.
    thread_id: Optional[str] = None
    # If applicable, the ID of the assistant that authored this message.
    assistant_id: Optional[str] = None
    # If applicable, the ID of the run associated with the authoring of this message.
    run_id: Optional[str] = None
    # A list of file IDs that the assistant should use.
    # Useful for tools like retrieval and code_interpreter that can access files.
    # A maximum of 10 files can be attached to a message.
    file_ids: Optional[List[Any]] = None
    # Files attached to this message.
    files: Optional[List[Any]] = None

    # Set of 16 key-value pairs that can be attached to an object.
    # This can be useful for storing additional information about the object in a structured format.
    # Keys can be a maximum of 64 characters long and values can be a maxium of 512 characters long.
    metadata: Optional[Dict[str, Any]] = None

    # The Unix timestamp (in seconds) for when the message was created.
    created_at: Optional[int] = None

    openai: Optional[OpenAI] = None
    openai_message: Optional[OpenAIThreadMessage] = None

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def to_dict(self) -> Dict[str, Any]:
        _dict: Dict[str, Any] = {
            "role": self.role,
            "content": self.content,
        }
        if self.file_ids:
            _dict["file_ids"] = self.file_ids
        if self.metadata:
            _dict["metadata"] = self.metadata
        return _dict

    @property
    def client(self) -> OpenAI:
        return self.openai or OpenAI()

    def load_from_openai(self, openai_message: OpenAIThreadMessage):
        self.id = openai_message.id
        self.object = openai_message.object
        self.role = openai_message.role
        self.content = openai_message.content
        self.created_at = openai_message.created_at

    def create(self) -> OpenAIThreadMessage:
        if self.thread_id is None:
            raise ThreadIdNotSet("Thread.id not set")

        request_body: Dict[str, Any] = {}
        if self.file_ids:
            _file_ids = self.file_ids or []
            # if self.files:
            #     for _file in self.files:
            #         _file_ids.append(_file.get_id())
            request_body["file_ids"] = _file_ids
        if self.metadata:
            request_body["metadata"] = self.metadata

        if not isinstance(self.content, str):
            raise TypeError("Message.content must be a string for create()")

        self.openai_message = self.client.beta.threads.messages.create(
            thread_id=self.thread_id, role="user", content=self.content, **request_body
        )
        self.load_from_openai(self.openai_message)
        return self.openai_message

    def get(self, use_cache: bool = True) -> OpenAIThreadMessage:
        if self.openai_message is not None and use_cache:
            return self.openai_message

        if self.thread_id is None:
            raise ThreadIdNotSet("Thread.id not set")

        _message_id = self.id or self.openai_message.id if self.openai_message else None
        if _message_id is not None:
            self.openai_message = self.client.beta.threads.messages.retrieve(
                thread_id=self.thread_id,
                message_id=_message_id,
            )
            self.load_from_openai(self.openai_message)
            return self.openai_message
        raise MessageIdNotSet("Message.id not set")

    def get_or_create(self, use_cache: bool = True) -> OpenAIThreadMessage:
        try:
            return self.get(use_cache=use_cache)
        except MessageIdNotSet:
            return self.create()

    def update(self) -> OpenAIThreadMessage:
        try:
            message_to_update = self.get()
            if message_to_update is not None:
                request_body: Dict[str, Any] = {}
                if self.metadata:
                    request_body["metadata"] = self.metadata

                self.openai_message = self.client.beta.threads.messages.update(
                    thread_id=message_to_update.thread_id,
                    message_id=message_to_update.id,
                    **request_body,
                )
                self.load_from_openai(self.openai_message)
                return self.openai_message
        except (ThreadIdNotSet, MessageIdNotSet):
            logger.warning("Message not available")
            raise



# from typing import Optional, Any, List
#
# try:
#     from sqlalchemy.dialects import postgresql
#     from sqlalchemy.engine import create_engine, Engine
#     from sqlalchemy.engine.row import Row
#     from sqlalchemy.inspection import inspect
#     from sqlalchemy.orm import Session, sessionmaker
#     from sqlalchemy.schema import MetaData, Table, Column
#     from sqlalchemy.sql.expression import text, select
#     from sqlalchemy.types import DateTime, String
# except ImportError:
#     raise ImportError("`sqlalchemy` not installed")
#
# from phi.conversation.schemas import ConversationRow
# from phi.conversation.storage.base import ConversationStorage
# from phi.utils.log import logger
#
#
# class PgConversationStorage(ConversationStorage):
#     def __init__(
#         self,
#         table_name: str,
#         schema: Optional[str] = "llm",
#         db_url: Optional[str] = None,
#         db_engine: Optional[Engine] = None,
#     ):
#         """
#         This class provides conversation storage using a postgres database.
#
#         The following order is used to determine the database connection:
#             1. Use the db_engine if provided
#             2. Use the db_url
#
#         :param table_name: The name of the table to store conversations in.
#         :param schema: The schema to store the table in.
#         :param db_url: The database URL to connect to.
#         :param db_engine: The database engine to use.
#         """
#         _engine: Optional[Engine] = db_engine
#         if _engine is None and db_url is not None:
#             _engine = create_engine(db_url)
#
#         if _engine is None:
#             raise ValueError("Must provide either db_url or db_engine")
#
#         # Database attributes
#         self.table_name: str = table_name
#         self.schema: Optional[str] = schema
#         self.db_url: Optional[str] = db_url
#         self.db_engine: Engine = _engine
#         self.metadata: MetaData = MetaData(schema=self.schema)
#
#         # Database session
#         self.Session: sessionmaker[Session] = sessionmaker(bind=self.db_engine)
#
#         # Database table for storage
#         self.table: Table = self.get_table()
#
#     def get_table(self) -> Table:
#         return Table(
#             self.table_name,
#             self.metadata,
#             # Primary key for this conversation.
#             Column("id", String, primary_key=True),
#             # Conversation name
#             Column("name", String),
#             # Name and type of user participating in this conversation.
#             Column("user_name", String),
#             Column("user_type", String),
#             # True if this conversation is active.
#             Column("is_active", postgresql.BOOLEAN, server_default=text("true")),
#             # -*- LLM data (name, model, etc.)
#             Column("llm", postgresql.JSONB),
#             # -*- Conversation memory
#             Column("memory", postgresql.JSONB),
#             # Metadata associated with this conversation.
#             Column("meta_data", postgresql.JSONB),
#             # Extra data associated with this conversation.
#             Column("extra_data", postgresql.JSONB),
#             # The timestamp of when this conversation was created.
#             Column("created_at", DateTime(timezone=True), server_default=text("now()")),
#             # The timestamp of when this conversation was last updated.
#             Column("updated_at", DateTime(timezone=True), onupdate=text("now()")),
#             extend_existing=True,
#         )
#
#     def table_exists(self) -> bool:
#         logger.debug(f"Checking if table exists: {self.table.name}")
#         try:
#             return inspect(self.db_engine).has_table(self.table.name, schema=self.schema)
#         except Exception as e:
#             logger.error(e)
#             return False
#
#     def create(self) -> None:
#         if not self.table_exists():
#             if self.schema is not None:
#                 with self.Session() as sess, sess.begin():
#                     logger.debug(f"Creating schema: {self.schema}")
#                     sess.execute(text(f"create schema if not exists {self.schema};"))
#             logger.debug(f"Creating table: {self.table_name}")
#             self.table.create(self.db_engine)
#
#     def _read(self, session: Session, conversation_id: str) -> Optional[Row[Any]]:
#         stmt = select(self.table).where(self.table.c.id == conversation_id)
#         try:
#             return session.execute(stmt).first()
#         except Exception:
#             # Create table if it does not exist
#             self.create()
#         return None
#
#     def read(self, conversation_id: str) -> Optional[ConversationRow]:
#         with self.Session() as sess, sess.begin():
#             existing_row: Optional[Row[Any]] = self._read(session=sess, conversation_id=conversation_id)
#             return ConversationRow.model_validate(existing_row) if existing_row is not None else None
#
#     def get_all_conversation_ids(self, user_name: Optional[str] = None) -> List[str]:
#         conversation_ids: List[str] = []
#         try:
#             with self.Session() as sess, sess.begin():
#                 # get all conversation ids for this user
#                 stmt = select(self.table)
#                 if user_name is not None:
#                     stmt = stmt.where(self.table.c.user_name == user_name)
#                 # order by created_at desc
#                 stmt = stmt.order_by(self.table.c.created_at.desc())
#                 # execute query
#                 rows = sess.execute(stmt).fetchall()
#                 for row in rows:
#                     if row is not None and row.id is not None:
#                         conversation_ids.append(row.id)
#         except Exception:
#             logger.debug(f"Table does not exist: {self.table.name}")
#         return conversation_ids
#
#     def get_all_conversations(self, user_name: Optional[str] = None) -> List[ConversationRow]:
#         conversations: List[ConversationRow] = []
#         try:
#             with self.Session() as sess, sess.begin():
#                 # get all conversation ids for this user
#                 stmt = select(self.table)
#                 if user_name is not None:
#                     stmt = stmt.where(self.table.c.user_name == user_name)
#                 # order by created_at desc
#                 stmt = stmt.order_by(self.table.c.created_at.desc())
#                 # execute query
#                 rows = sess.execute(stmt).fetchall()
#                 for row in rows:
#                     if row.id is not None:
#                         conversations.append(ConversationRow.model_validate(row))
#         except Exception:
#             logger.debug(f"Table does not exist: {self.table.name}")
#         return conversations
#
#     def upsert(self, conversation: ConversationRow) -> Optional[ConversationRow]:
#         """
#         Create a new conversation if it does not exist, otherwise update the existing conversation.
#         """
#
#         with self.Session() as sess, sess.begin():
#             # Create an insert statement
#             stmt = postgresql.insert(self.table).values(
#                 id=conversation.id,
#                 name=conversation.name,
#                 user_name=conversation.user_name,
#                 user_type=conversation.user_type,
#                 is_active=conversation.is_active,
#                 llm=conversation.llm,
#                 memory=conversation.memory,
#                 meta_data=conversation.meta_data,
#                 extra_data=conversation.extra_data,
#             )
#
#             # Define the upsert if the id already exists
#             # See: https://docs.sqlalchemy.org/en/20/dialects/postgresql.html#postgresql-insert-on-conflict
#             stmt = stmt.on_conflict_do_update(
#                 index_elements=["id"],
#                 set_=dict(
#                     name=conversation.name,
#                     user_name=conversation.user_name,
#                     user_type=conversation.user_type,
#                     is_active=conversation.is_active,
#                     llm=conversation.llm,
#                     memory=conversation.memory,
#                     meta_data=conversation.meta_data,
#                     extra_data=conversation.extra_data,
#                 ),  # The updated value for each column
#             )
#
#             try:
#                 sess.execute(stmt)
#             except Exception:
#                 # Create table and try again
#                 self.create()
#                 sess.execute(stmt)
#         return self.read(conversation_id=conversation.id)
#
#     def end(self, conversation_id: str) -> None:
#         with self.Session() as sess, sess.begin():
#             # Check if conversation exists in the database
#             existing_row: Optional[Row[Any]] = self._read(session=sess, conversation_id=conversation_id)
#             # If conversation exists, set is_active to False
#             if existing_row is not None:
#                 stmt = self.table.update().where(self.table.c.id == conversation_id).values(is_active=False)
#                 sess.execute(stmt)
#
#     def delete(self) -> None:
#         if self.table_exists():
#             logger.debug(f"Deleting table: {self.table_name}")
#             self.table.drop(self.db_engine)



from typing import Any, Optional, Dict, List

from pydantic import BaseModel, ConfigDict

from phi.assistant.run import Run
from phi.assistant.message import Message
from phi.assistant.exceptions import AssistantIdNotSet, ThreadIdNotSet
from phi.utils.log import logger

try:
    from openai import OpenAI
    from openai.types.beta.assistant import Assistant as OpenAIAssistant
    from openai.types.beta.thread import Thread as OpenAIThread
    from openai.types.beta.thread_deleted import ThreadDeleted as OpenAIThreadDeleted
except ImportError:
    logger.error("`openai` not installed")
    raise


class Thread(BaseModel):
    # -*- Thread settings
    # Thread id which can be referenced in API endpoints.
    id: Optional[str] = None
    # The object type, populated by the API. Always thread.
    object: Optional[str] = None

    # A list of messages to start the thread with.
    messages: Optional[List[Message | Dict]] = None

    # The ID of the assistant for this message.
    assistant_id: Optional[str] = None

    # Set of 16 key-value pairs that can be attached to an object.
    # This can be useful for storing additional information about the object in a structured format.
    # Keys can be a maximum of 64 characters long and values can be a maxium of 512 characters long.
    metadata: Optional[Dict[str, Any]] = None

    # True if this thread is active
    is_active: bool = True
    # The Unix timestamp (in seconds) for when the thread was created.
    created_at: Optional[int] = None

    openai: Optional[OpenAI] = None
    openai_thread: Optional[OpenAIThread] = None
    openai_assistant: Optional[OpenAIAssistant] = None

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @property
    def client(self) -> OpenAI:
        return self.openai or OpenAI()

    def load_from_openai(self, openai_thread: OpenAIThread):
        self.id = openai_thread.id
        self.object = openai_thread.object
        self.created_at = openai_thread.created_at

    def create(self) -> OpenAIThread:
        request_body: Dict[str, Any] = {}
        if self.messages:
            _messages = []
            for _message in self.messages:
                if isinstance(_message, Message):
                    _messages.append(_message.to_dict())
                else:
                    _messages.append(_message)
            request_body["messages"] = _messages
        if self.metadata:
            request_body["metadata"] = self.metadata

        self.openai_thread = self.client.beta.threads.create(**request_body)
        self.load_from_openai(self.openai_thread)
        return self.openai_thread

    def get(self, use_cache: bool = True) -> OpenAIThread:
        if self.openai_thread is not None and use_cache:
            return self.openai_thread

        _thread_id = self.id or self.openai_thread.id if self.openai_thread else None
        if _thread_id is not None:
            self.openai_thread = self.client.beta.threads.retrieve(
                thread_id=_thread_id,
            )
            self.load_from_openai(self.openai_thread)
            return self.openai_thread
        raise ThreadIdNotSet("Thread.id not set")

    def get_or_create(self, use_cache: bool = True) -> OpenAIThread:
        try:
            return self.get(use_cache=use_cache)
        except ThreadIdNotSet:
            return self.create()

    def update(self) -> OpenAIThread:
        try:
            thread_to_update = self.get()
            if thread_to_update is not None:
                request_body: Dict[str, Any] = {}
                if self.metadata:
                    request_body["metadata"] = self.metadata

                self.openai_thread = self.client.beta.threads.update(
                    thread_id=thread_to_update.id,
                    **request_body,
                )
                self.load_from_openai(self.openai_thread)
                return self.openai_thread
        except ThreadIdNotSet:
            logger.warning("Thread not available")
            raise

    def delete(self) -> OpenAIThreadDeleted:
        try:
            thread_to_delete = self.get()
            if thread_to_delete is not None:
                deletion_status = self.client.beta.threads.delete(
                    thread_id=thread_to_delete.id,
                )
                return deletion_status
        except ThreadIdNotSet:
            logger.warning("Thread not available")
            raise

    def create_message(self, message: Optional[Message] = None, **kwargs) -> Message:
        try:
            if message is None:
                message = Message(**kwargs)
        except Exception as e:
            logger.error(f"Error creating message: {e}")
            raise
        message.thread_id = self.id

        logger.debug(f"Creating message: {message}")
        message.create()
        return message

    def create_run(
        self, assistant_id: Optional[str] = None, run: Optional[Run] = None, use_cache: bool = True, **kwargs
    ) -> Run:
        try:
            thread_to_run = self.get(use_cache=use_cache)
        except ThreadIdNotSet:
            logger.warning("Thread not available")
            raise

        if assistant_id is None:
            assistant_id = self.assistant_id
        if assistant_id is None:
            raise AssistantIdNotSet("Assistant.id not set")

        try:
            if run is None:
                run = Run(**kwargs)
        except Exception as e:
            logger.error(f"Error creating run: {e}")
            raise

        run.thread_id = thread_to_run.id
        run.assistant_id = assistant_id

        logger.debug(f"Creating run: {run}")
        run.create()
        return run

    def get_messages(self, use_cache: bool = True) -> List[Message]:
        try:
            thread_to_read = self.get(use_cache=use_cache)
        except ThreadIdNotSet:
            logger.warning("Thread not available")
            raise

        thread_messages = self.client.beta.threads.messages.list(
            thread_id=thread_to_read.id,
        )
        return [Message(**message.model_dump()) for message in thread_messages]



from os import getenv
from typing import Union, Dict, List

from httpx import Response

from phi.api.api import api, invalid_response
from phi.api.routes import ApiRoutes
from phi.api.schemas.conversation import (
    ConversationWorkspace,
    ConversationEventCreate,
    ConversationMonitorCreate,
)
from phi.constants import WORKSPACE_ID_ENV_VAR, WORKSPACE_HASH_ENV_VAR, WORKSPACE_KEY_ENV_VAR
from phi.cli.settings import phi_cli_settings
from phi.utils.common import str_to_int
from phi.utils.log import logger


def create_conversation_monitor(monitor: ConversationMonitorCreate) -> bool:
    if not phi_cli_settings.api_enabled:
        return True

    # logger.debug("--o-o-- Creating Conversation Monitor")
    with api.AuthenticatedClient() as api_client:
        try:
            conversation_workspace = ConversationWorkspace(
                id_workspace=str_to_int(getenv(WORKSPACE_ID_ENV_VAR)),
                ws_hash=getenv(WORKSPACE_HASH_ENV_VAR),
                ws_key=getenv(WORKSPACE_KEY_ENV_VAR),
            )
            r: Response = api_client.post(
                ApiRoutes.CONVERSATION_MONITOR_CREATE,
                json={
                    "monitor": monitor.model_dump(exclude_none=True),
                    "workspace": conversation_workspace.model_dump(exclude_none=True),
                },
            )
            if invalid_response(r):
                return False

            response_json: Union[Dict, List] = r.json()
            if response_json is None:
                return False

            # logger.debug(f"Response: {response_json}")
            return True
        except Exception as e:
            logger.debug(f"Could not create conversation monitor: {e}")
    return False


def create_conversation_event(conversation: ConversationEventCreate) -> bool:
    if not phi_cli_settings.api_enabled:
        return True

    # logger.debug("--o-o-- Creating Conversation Event")
    with api.AuthenticatedClient() as api_client:
        try:
            conversation_workspace = ConversationWorkspace(
                id_workspace=str_to_int(getenv(WORKSPACE_ID_ENV_VAR)),
                ws_hash=getenv(WORKSPACE_HASH_ENV_VAR),
                ws_key=getenv(WORKSPACE_KEY_ENV_VAR),
            )
            r: Response = api_client.post(
                ApiRoutes.CONVERSATION_EVENT_CREATE,
                json={
                    "event": conversation.model_dump(exclude_none=True),
                    "workspace": conversation_workspace.model_dump(exclude_none=True),
                },
            )
            if invalid_response(r):
                return False

            response_json: Union[Dict, List] = r.json()
            if response_json is None:
                return False

            # logger.debug(f"Response: {response_json}")
            return True
        except Exception as e:
            logger.debug(f"Could not log conversation event: {e}")
    return False



from typing import List, Any, Optional, Dict

from pydantic import BaseModel, ConfigDict, field_validator

from phi.assistant.tool import Tool
from phi.assistant.thread import Thread
from phi.assistant.message import Message
from phi.assistant.row import AssistantRow
from phi.assistant.storage import AssistantStorage
from phi.assistant.exceptions import AssistantIdNotSet
from phi.knowledge.base import KnowledgeBase

from phi.utils.log import logger, set_log_level_to_debug

try:
    from openai import OpenAI
    from openai.types.beta.assistant import Assistant as OpenAIAssistant
    from openai.types.beta.asssitant_deleted import AsssitantDeleted as OpenAIAssistantDeleted
except ImportError:
    logger.error("`openai` not installed")
    raise


class Assistant(BaseModel):
    # -*- LLM settings
    model: str = "gpt-4-1106-preview"
    openai: Optional[OpenAI] = None

    # -*- Assistant settings
    # Assistant id which can be referenced in API endpoints.
    id: Optional[str] = None
    # The object type, populated by the API. Always assistant.
    object: Optional[str] = None
    # The name of the assistant. The maximum length is 256 characters.
    name: Optional[str] = None
    # The description of the assistant. The maximum length is 512 characters.
    description: Optional[str] = None
    # The system instructions that the assistant uses. The maximum length is 32768 characters.
    instructions: Optional[str] = None

    # -*- Assistant Tools
    # A list of tool enabled on the assistant. There can be a maximum of 128 tools per assistant.
    # Tools can be of types code_interpreter, retrieval, or function.
    tools: Optional[List[Tool | Dict]] = None

    # -*- Assistant Files
    # A list of file IDs attached to this assistant.
    # There can be a maximum of 20 files attached to the assistant.
    # Files are ordered by their creation date in ascending order.
    file_ids: Optional[List[Any]] = None
    # Files attached to this assistant.
    files: Optional[List[Any]] = None

    # -*- Assistant Storage
    storage: Optional[AssistantStorage] = None
    # Create table if it doesn't exist
    create_storage: bool = True
    # AssistantRow from the database: DO NOT SET THIS MANUALLY
    database_row: Optional[AssistantRow] = None

    # -*- Assistant Knowledge Base
    knowledge_base: Optional[KnowledgeBase] = None

    # -*- Thread settings
    # Add an introduction message to the thread
    introduction: Optional[str] = None

    # Set of 16 key-value pairs that can be attached to an object.
    # This can be useful for storing additional information about the object in a structured format.
    # Keys can be a maximum of 64 characters long and values can be a maxium of 512 characters long.
    metadata: Optional[Dict[str, Any]] = None

    # True if this assistant is active
    is_active: bool = True
    # The Unix timestamp (in seconds) for when the assistant was created.
    created_at: Optional[int] = None

    # If True, show debug logs
    debug_mode: bool = False
    # Enable monitoring on phidata.com
    monitoring: bool = False

    openai_assistant: Optional[OpenAIAssistant] = None

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @field_validator("debug_mode", mode="before")
    def set_log_level(cls, v: bool) -> bool:
        if v:
            set_log_level_to_debug()
            logger.debug("Debug logs enabled")
        return v

    @property
    def client(self) -> OpenAI:
        return self.openai or OpenAI()

    def load_from_openai(self, openai_assistant: OpenAIAssistant):
        self.id = openai_assistant.id
        self.object = openai_assistant.object
        self.created_at = openai_assistant.created_at

    def create(self) -> OpenAIAssistant:
        request_body: Dict[str, Any] = {}
        if self.name:
            request_body["name"] = self.name
        if self.description:
            request_body["description"] = self.description
        if self.instructions:
            request_body["instructions"] = self.instructions
        if self.tools:
            _tools = []
            for _tool in self.tools:
                if isinstance(_tool, Tool):
                    _tools.append(_tool.to_dict())
                else:
                    _tools.append(_tool)
            request_body["tools"] = _tools
        if self.file_ids or self.files:
            _file_ids = self.file_ids or []
            # if self.files:
            #     for _file in self.files:
            #         _file_ids.append(_file.get_id())
            request_body["file_ids"] = _file_ids
        if self.metadata:
            request_body["metadata"] = self.metadata

        self.openai_assistant = self.client.beta.assistants.create(
            model=self.model,
            **request_body,
        )
        self.load_from_openai(self.openai_assistant)
        return self.openai_assistant

    def get(self, use_cache: bool = True) -> OpenAIAssistant:
        if self.openai_assistant is not None and use_cache:
            return self.openai_assistant

        _assistant_id = self.id or self.openai_assistant.id if self.openai_assistant else None
        if _assistant_id is not None:
            self.openai_assistant = self.client.beta.assistants.retrieve(
                assistant_id=_assistant_id,
            )
            self.load_from_openai(self.openai_assistant)
            return self.openai_assistant
        raise AssistantIdNotSet("Assistant.id not set")

    def get_or_create(self, use_cache: bool = True) -> OpenAIAssistant:
        try:
            return self.get(use_cache=use_cache)
        except AssistantIdNotSet:
            return self.create()

    def update(self) -> OpenAIAssistant:
        try:
            assistant_to_update = self.get()
            if assistant_to_update is not None:
                request_body: Dict[str, Any] = {}
                if self.name:
                    request_body["name"] = self.name
                if self.description:
                    request_body["description"] = self.description
                if self.instructions:
                    request_body["instructions"] = self.instructions
                if self.tools:
                    _tools = []
                    for _tool in self.tools:
                        if isinstance(_tool, Tool):
                            _tools.append(_tool.to_dict())
                        else:
                            _tools.append(_tool)
                    request_body["tools"] = _tools
                if self.file_ids or self.files:
                    _file_ids = self.file_ids or []
                    # if self.files:
                    #     for _file in self.files:
                    #         _file_ids.append(_file.get_id())
                    request_body["file_ids"] = _file_ids
                if self.metadata:
                    request_body["metadata"] = self.metadata

                self.openai_assistant = self.client.beta.assistants.update(
                    assistant_id=assistant_to_update.id,
                    model=self.model,
                    **request_body,
                )
                self.load_from_openai(self.openai_assistant)
                return self.openai_assistant
        except AssistantIdNotSet:
            logger.warning("Assistant not available")
            raise

    def delete(self) -> OpenAIAssistantDeleted:
        try:
            assistant_to_delete = self.get()
            if assistant_to_delete is not None:
                deletion_status = self.client.beta.assistants.delete(
                    assistant_id=assistant_to_delete.id,
                )
                return deletion_status
        except AssistantIdNotSet:
            logger.warning("Assistant not available")
            raise

    def create_thread(self, thread: Optional[Thread] = None) -> Thread:
        try:
            assistant = self.get()
        except AssistantIdNotSet:
            logger.warning("Assistant not available")
            raise

        if thread is None:
            thread = Thread()
            if self.introduction:
                thread.messages = [
                    Message(
                        role="assistant",
                        content=self.introduction,
                    )
                ]

        thread.assistant_id = assistant.id
        thread.openai_assistant = assistant

        logger.debug(f"Creating thread: {thread}")
        thread.create()
        return thread



from typing import Any, Optional, Dict, List

from pydantic import BaseModel, ConfigDict

from phi.assistant.tool import Tool
from phi.assistant.exceptions import ThreadIdNotSet, AssistantIdNotSet, RunIdNotSet
from phi.utils.log import logger

try:
    from openai import OpenAI
    from openai.types.beta.threads.run import Run as OpenAIRun
except ImportError:
    logger.error("`openai` not installed")
    raise


class Run(BaseModel):
    # -*- Run settings
    # Run id which can be referenced in API endpoints.
    id: Optional[str] = None
    # The object type, populated by the API. Always assistant.run.
    object: Optional[str] = None

    # The ID of the thread that was executed on as a part of this run.
    thread_id: Optional[str] = None
    # The ID of the assistant used for execution of this run.
    assistant_id: Optional[str] = None

    # The status of the run, which can be either
    # queued, in_progress, requires_action, cancelling, cancelled, failed, completed, or expired.
    status: Optional[str] = None

    # Details on the action required to continue the run. Will be null if no action is required.
    required_action: Optional[Dict[str, Any]] = None

    # The last error associated with this run. Will be null if there are no errors.
    last_error: Optional[Dict[str, Any]] = None

    # True if this run is active
    is_active: bool = True
    # The Unix timestamp (in seconds) for when the run was created.
    created_at: Optional[int] = None
    # The Unix timestamp (in seconds) for when the run will expire.
    expires_at: Optional[int] = None
    # The Unix timestamp (in seconds) for when the run was started.
    started_at: Optional[int] = None
    # The Unix timestamp (in seconds) for when the run was cancelled.
    cancelled_at: Optional[int] = None
    # The Unix timestamp (in seconds) for when the run failed.
    failed_at: Optional[int] = None
    # The Unix timestamp (in seconds) for when the run was completed.
    completed_at: Optional[int] = None
    # The list of File IDs the assistant used for this run.
    file_ids: Optional[List[Any]] = None

    # The ID of the Model to be used to execute this run. If a value is provided here,
    # it will override the model associated with the assistant.
    # If not, the model associated with the assistant will be used.
    model: Optional[str] = None
    # Override the default system message of the assistant.
    # This is useful for modifying the behavior on a per-run basis.
    instructions: Optional[str] = None
    # Override the tools the assistant can use for this run.
    # This is useful for modifying the behavior on a per-run basis.
    tools: Optional[List[Tool | Dict]] = None

    # Set of 16 key-value pairs that can be attached to an object.
    # This can be useful for storing additional information about the object in a structured format.
    # Keys can be a maximum of 64 characters long and values can be a maxium of 512 characters long.
    metadata: Optional[Dict[str, Any]] = None

    # If True, show debug logs
    debug_mode: bool = False
    # Enable monitoring on phidata.com
    monitoring: bool = False

    openai: Optional[OpenAI] = None
    openai_run: Optional[OpenAIRun] = None

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @property
    def client(self) -> OpenAI:
        return self.openai or OpenAI()

    def load_from_openai(self, openai_run: OpenAIRun):
        self.id = openai_run.id
        self.object = openai_run.object
        self.created_at = openai_run.created_at

    def create(self) -> OpenAIRun:
        if self.thread_id is None:
            raise ThreadIdNotSet("Thread.id not set")

        if self.assistant_id is None:
            raise AssistantIdNotSet("Assistant.id not set")

        request_body: Dict[str, Any] = {}
        if self.model:
            request_body["model"] = self.model
        if self.instructions:
            request_body["instructions"] = self.instructions
        if self.tools:
            _tools = []
            for _tool in self.tools:
                if isinstance(_tool, Tool):
                    _tools.append(_tool.to_dict())
                else:
                    _tools.append(_tool)
            request_body["tools"] = _tools
        if self.metadata:
            request_body["metadata"] = self.metadata

        self.openai_run = self.client.beta.threads.runs.create(
            thread_id=self.thread_id, assistant_id=self.assistant_id, **request_body
        )
        self.load_from_openai(self.openai_run)
        return self.openai_run

    def get(self, use_cache: bool = True) -> OpenAIRun:
        if self.openai_run is not None and use_cache:
            return self.openai_run

        if self.thread_id is None:
            raise ThreadIdNotSet("Thread.id not set")

        _run_id = self.id or self.openai_run.id if self.openai_run else None
        if _run_id is not None:
            self.openai_run = self.client.beta.threads.runs.retrieve(
                thread_id=self.thread_id,
                run_id=_run_id,
            )
            self.load_from_openai(self.openai_run)
            return self.openai_run
        raise RunIdNotSet("Message.id not set")

    def get_or_create(self, use_cache: bool = True) -> OpenAIRun:
        try:
            return self.get(use_cache=use_cache)
        except RunIdNotSet:
            return self.create()

    def update(self) -> OpenAIRun:
        try:
            run_to_update = self.get()
            if run_to_update is not None:
                request_body: Dict[str, Any] = {}
                if self.metadata:
                    request_body["metadata"] = self.metadata

                self.openai_run = self.client.beta.threads.runs.update(
                    thread_id=run_to_update.thread_id,
                    run_id=run_to_update.id,
                    **request_body,
                )
                self.load_from_openai(self.openai_run)
                return self.openai_run
        except (ThreadIdNotSet, RunIdNotSet):
            logger.warning("Message not available")
            raise

    def wait_for_completion(self, timeout: Optional[int] = None) -> OpenAIRun:
        import time

        start_time = time.time()
        while True:
            logger.debug(f"Waiting for run {self.id} to complete")
            run = self.get(use_cache=False)
            logger.debug(f"Run {run.id}: {run}")
            logger.debug(f"Run {run.id} status: {run.status}")
            if run.status == "completed":
                return run
            if timeout is not None and time.time() - start_time > timeout:
                raise TimeoutError(f"Run {run.id} did not complete within {timeout} seconds")
            time.sleep(1)



from typing import Any, Dict
from pydantic import BaseModel


class BaseTool(BaseModel):
    """Model for LLM Tools"""

    # The type of tool
    type: str

    def to_dict(self) -> Dict[str, Any]:
        return self.model_dump(exclude_none=True)
